"""Benchmark utilities."""

from matfree.backend import func, linalg, np, time


def rmse_relative(received, *, expected):
    """Compute the relative root-mean-square error."""
    return linalg.vector_norm((received - expected) / expected) / np.sqrt(expected.size)


def error_and_time(fun, error_fun):
    """Compute error and runtime of a function with a single outputs."""

    @func.wraps(fun)
    def fun_wrapped(*args, **kwargs):
        # Execute once for compilation
        _ = fun(*args, **kwargs)

        # Execute and time
        t0 = time.perf_counter()
        result = fun(*args, **kwargs)
        result.block_until_ready()
        t1 = time.perf_counter()
        return error_fun(result), (t1 - t0)

    return fun_wrapped



"""Monte-Carlo estimation."""

from matfree.backend import containers, control_flow, func, np, prng
from matfree.backend.typing import Array, Callable


def estimate(
    fun: Callable,
    /,
    *,
    key: Array,
    sample_fun: Callable,
    num_batches: int = 1,
    num_samples_per_batch: int = 10_000,
    target_fun=np.nanmean,
) -> Array:
    """Monte-Carlo estimation: Compute the expected value of a function.

    Parameters
    ----------
    fun:
        Function whose expected value shall be estimated.
    key:
        Pseudo-random number generator key.
    sample_fun:
        Sampling function.
        For trace-estimation, use
        either [montecarlo.normal(...)][matfree.montecarlo.normal]
        or [montecarlo.rademacher(...)][matfree.montecarlo.normal].
    num_batches:
        Number of batches when computing arithmetic means.
    num_samples_per_batch:
        Number of samples per batch.
    target_fun:
        The target function to compute. Usually, this is np.mean. But any other
        statistical function with a signature like
        [those](https://data-apis.org/array-api/2022.12/API_specification/statistical_functions.html)
        would work. The default is np.nanmean.
    """
    fun_mc = _montecarlo(fun, sample_fun=sample_fun)
    fun_single_batch = _stats_via_vmap(fun_mc, num_samples_per_batch, target_fun)
    fun_batched = _stats_via_map(fun_single_batch, num_batches, target_fun)
    return fun_batched(key)


def _montecarlo(f, /, *, sample_fun):
    """Turn a function into a Monte-Carlo problem.

    More specifically, f(x) becomes g(key) = f(h(key)),
    using a sample function h: key -> x.
    This can then be evaluated and averaged in batches, loops, and compositions thereof.
    """
    # todo: what about randomised QMC? How do we best implement this?

    def f_mc(key, /):
        sample = sample_fun(key)
        return f(sample)

    return f_mc


def _stats_via_vmap(f, num, /, target_fun):
    """Compute summary statistics via jax.vmap."""

    def f_mean(key, /):
        subkeys = prng.split(key, num)
        fx_values = func.vmap(f)(subkeys)
        return target_fun(fx_values, axis=0)

    return f_mean


def _stats_via_map(f, num, /, target_fun):
    """Compute summary statistics via jax.lax.map."""

    def f_mean(key, /):
        subkeys = prng.split(key, num)
        fx_values = control_flow.map(f, subkeys)
        return target_fun(fx_values, axis=0)

    return f_mean


def normal(*, shape, dtype=float):
    """Construct a function that samples from a standard normal distribution."""

    def fun(key):
        return prng.normal(key, shape=shape, dtype=dtype)

    return fun


def rademacher(*, shape, dtype=float):
    """Construct a function that samples from a Rademacher distribution."""

    def fun(key):
        return prng.rademacher(key, shape=shape, dtype=dtype)

    return fun


class _VDCState(containers.NamedTuple):
    n: int
    vdc: float
    denom: int


def van_der_corput(n, /, base=2):
    """Compute the 'n'th element of the Van-der-Corput sequence."""
    state = _VDCState(n, vdc=0, denom=1)

    vdc_modify = func.partial(_van_der_corput_modify, base=base)
    state = control_flow.while_loop(_van_der_corput_cond, vdc_modify, state)
    return state.vdc


def _van_der_corput_cond(state: _VDCState):
    return state.n > 0


def _van_der_corput_modify(state: _VDCState, *, base):
    denom = state.denom * base
    num, remainder = divmod(state.n, base)
    vdc = state.vdc + remainder / denom
    return _VDCState(num, vdc, denom)



"""Matrix decomposition algorithms."""

from matfree.backend import containers, control_flow, func, linalg, np
from matfree.backend.typing import Any, Array, Callable, Tuple


def svd(v0: Array, depth: int, Av: Callable, vA: Callable, matrix_shape: Tuple[int]):
    """Approximate singular value decomposition.

    Parameters
    ----------
    v0:
        Initial vector for Golub-Kahan-Lanczos bidiagonalisation.
    depth:
        Depth of the Krylov space constructed by Golub-Kahan-Lanczos bidiagonalisation.
        Choosing `depth = min(nrows, ncols) - 1` would yield behaviour similar to
        e.g. `np.linalg.svd`.
    Av:
        Matrix-vector product function.
    vA:
        Vector-matrix product function.
    matrix_shape:
        Shape of the matrix involved in matrix-vector and vector-matrix products.
    """
    # Factorise the matrix
    alg = golub_kahan_lanczos_bidiagonal(depth, matrix_shape=matrix_shape)
    u, (d, e), vt, _ = decompose_fori_loop(0, depth + 1, v0, Av, vA, alg=alg)

    # Compute SVD of factorisation
    B = _bidiagonal_dense(d, e)
    U, S, Vt = linalg.svd(B, full_matrices=False)

    # Combine orthogonal transformations
    return u @ U, S, Vt @ vt


def _bidiagonal_dense(d, e):
    diag = linalg.diagonal_matrix(d)
    offdiag = linalg.diagonal_matrix(e, 1)
    return diag + offdiag


class DecompAlg(containers.NamedTuple):
    """Matrix decomposition algorithm."""

    init: Callable
    """Initialise the state of the algorithm. Usually, this involves pre-allocation."""

    step: Callable
    """Compute the next iteration."""

    extract: Callable
    """Extract the solution from the state of the algorithm."""


# all arguments are positional-only because we will rename arguments a lot
def decompose_fori_loop(lower, upper, v0, *matvec_funs, alg: DecompAlg):
    r"""Decompose a matrix purely based on matvec-products with A.

    This behaviour of this function is equivalent to

    ```python
    def decompose(lower, upper, v0, *matvec_funs, alg):
        state = alg.init(v0)
        for _ in range(lower, upper):
            state = alg.step(state, *matvec_funs)
        return alg.extract(state)
    ```

    but the implementation uses JAX' fori_loop.
    """
    # todo: turn the "practically equivalent" bit above into a doctest.
    init_val = alg.init(v0)

    def body_fun(_, s):
        return alg.step(s, *matvec_funs)

    result = control_flow.fori_loop(lower, upper, body_fun=body_fun, init_val=init_val)
    return alg.extract(result)


def lanczos_tridiagonal(depth, /) -> DecompAlg:
    """**Lanczos** algorithm with pre-allocation and re-orthogonalisation.

    Decompose a matrix into a product of orthogonal-**tridiagonal**-orthogonal matrices.
    Use this algorithm for approximate **eigenvalue** decompositions.

    Lanczos' algorithm assumes a symmetric matrix.
    """
    # this algorithm is massively unstable.
    # but despite this instability, quadrature might be stable?
    # https://www.sciencedirect.com/science/article/abs/pii/S0920563200918164
    return DecompAlg(
        init=func.partial(_lanczos_tridiagonal_init, depth),
        step=_lanczos_tridiagonal_apply,
        extract=_lanczos_tridiagonal_extract,
    )


class _LanczosState(containers.NamedTuple):
    i: int
    basis: Any
    tridiag: Any
    q: Any


def _lanczos_tridiagonal_init(depth: int, init_vec: Array) -> _LanczosState:
    (ncols,) = np.shape(init_vec)
    if depth >= ncols or depth < 1:
        raise ValueError

    diag = np.zeros((depth + 1,))
    offdiag = np.zeros((depth,))
    basis = np.zeros((depth + 1, ncols))

    return _LanczosState(0, basis, (diag, offdiag), init_vec)


def _lanczos_tridiagonal_apply(state: _LanczosState, Av: Callable) -> _LanczosState:
    i, basis, (diag, offdiag), vec = state

    # This one is a hack:
    #
    # Re-orthogonalise against ALL basis elements before storing.
    # Note: we re-orthogonalise against ALL columns of Q, not just
    # the ones we have already computed. This increases the complexity
    # of the whole iteration from n(n+1)/2 to n^2, but has the advantage
    # that the whole computation has static bounds (thus we can JIT it all).
    # Since 'Q' is padded with zeros, the numerical values are identical
    # between both modes of computing.
    vec, length = _normalise(vec)
    vec, _ = _gram_schmidt_orthogonalise_set(vec, basis)

    # I don't know why, but this re-normalisation is soooo crucial
    vec, _ = _normalise(vec)
    basis = basis.at[i, :].set(vec)

    # When i==0, Q[i-1] is Q[-1] and again, we benefit from the fact
    #  that Q is initialised with zeros.
    vec = Av(vec)
    basis_vectors_previous = np.asarray([basis[i], basis[i - 1]])
    vec, (coeff, _) = _gram_schmidt_orthogonalise_set(vec, basis_vectors_previous)
    diag = diag.at[i].set(coeff)
    offdiag = offdiag.at[i - 1].set(length)

    return _LanczosState(i + 1, basis, (diag, offdiag), vec)


def _lanczos_tridiagonal_extract(state: _LanczosState, /):
    _, basis, (diag, offdiag), _ = state
    return basis, (diag, offdiag)


def golub_kahan_lanczos_bidiagonal(depth, /, matrix_shape) -> DecompAlg:
    """**Golub-Kahan-Lanczos** algorithm with pre-allocation and re-orthogonalisation.

    Decompose a matrix into a product of orthogonal-**bidiagonal**-orthogonal matrices.
    Use this algorithm for approximate **singular value** decompositions.
    """
    return DecompAlg(
        init=func.partial(_gkl_bidiagonal_init, depth, matrix_shape),
        step=_gkl_bidiagonal_apply,
        extract=_gkl_bidiagonal_extract,
    )


class _GKLState(containers.NamedTuple):
    i: int
    Us: Any
    Vs: Any
    alphas: Any
    betas: Any
    beta: Any
    vk: Any


def _gkl_bidiagonal_init(depth: int, matrix_shape, init_vec: Array) -> _GKLState:
    nrows, ncols = matrix_shape
    alphas = np.zeros((depth + 1,))
    betas = np.zeros((depth + 1,))
    Us = np.zeros((depth + 1, nrows))
    Vs = np.zeros((depth + 1, ncols))
    v0, _ = _normalise(init_vec)
    return _GKLState(0, Us, Vs, alphas, betas, 0.0, v0)


def _gkl_bidiagonal_apply(state: _GKLState, Av: Callable, vA: Callable) -> _GKLState:
    i, Us, Vs, alphas, betas, beta, vk = state
    Vs = Vs.at[i].set(vk)
    betas = betas.at[i].set(beta)

    uk = Av(vk) - beta * Us[i - 1]
    uk, alpha = _normalise(uk)
    uk, _ = _gram_schmidt_orthogonalise_set(uk, Us)  # hack
    uk, _ = _normalise(uk)  # hack
    Us = Us.at[i].set(uk)
    alphas = alphas.at[i].set(alpha)

    vk = vA(uk) - alpha * vk
    vk, beta = _normalise(vk)
    vk, _ = _gram_schmidt_orthogonalise_set(vk, Vs)  # hack
    vk, _ = _normalise(vk)  # hack

    return _GKLState(i + 1, Us, Vs, alphas, betas, beta, vk)


def _gkl_bidiagonal_extract(state: _GKLState, /):
    _, uk_all, vk_all, alphas, betas, beta, vk = state
    return uk_all.T, (alphas, betas[1:]), vk_all, (beta, vk)


def _normalise(vec):
    length = linalg.vector_norm(vec)
    return vec / length, length


def _gram_schmidt_orthogonalise_set(vec, vectors):  # Gram-Schmidt
    vec, coeffs = control_flow.scan(_gram_schmidt_orthogonalise, vec, xs=vectors)
    return vec, coeffs


def _gram_schmidt_orthogonalise(vec1, vec2):
    coeff = linalg.vecdot(vec1, vec2)
    vec_ortho = vec1 - coeff * vec2
    return vec_ortho, coeff



"""Tests for Monte-Carlo machinery."""

from matfree import montecarlo
from matfree.backend import np, prng, testing


@testing.parametrize("key", [prng.prng_key(1)])
@testing.parametrize("num_batches, num_samples", [[1, 10_000], [10_000, 1], [100, 100]])
def test_mean(key, num_batches, num_samples):
    """Assert that the mean estimate is accurate."""

    def fun(x):
        return x**2

    received = montecarlo.estimate(
        fun,
        num_batches=num_batches,
        num_samples_per_batch=num_samples,
        key=key,
        sample_fun=montecarlo.normal(shape=()),
    )
    assert np.allclose(received, 1.0, rtol=1e-1)



"""What is the fastest way of computing trace(A^5)."""
from matfree import benchmark_util, hutch, montecarlo, slq
from matfree.backend import func, linalg, np, plt, prng
from matfree.backend.progressbar import progressbar


def problem(n):
    """Create an example problem."""

    # This function has a Jacobian with x-shaped sparsity pattern
    # We expect control variates to do pretty well
    # (But I don't know why)
    def f(x):
        return np.sin(np.roll(np.sin(np.flip(np.cos(x)) + 1) ** 2, 2)) * np.sin(x**2)

    key = prng.prng_key(seed=2)
    x0 = prng.uniform(key, shape=(n,))

    _, jvp = func.linearize(f, x0)
    J = func.jacfwd(f)(x0)
    A = J @ J @ J @ J
    trace = linalg.trace(A)
    sample_fun = montecarlo.normal(shape=(n,), dtype=float)

    def Av(v):
        return jvp(jvp(jvp(jvp(v))))

    return (lambda x: x**4, jvp, Av, trace, A), (key, sample_fun)


def evaluate_all(fun, outer_loop, inner_loop):
    """Evaluate all metrics of a function."""
    errors, stds, times = [], [], []
    for n in outer_loop:
        fun_bind = func.partial(fun, int(n))
        err, tim = evaluate(fun_bind, inner_loop)

        errors.append(np.mean(err))
        stds.append(np.std(err))
        times.append(np.amin(tim))

    return np.asarray(errors), np.asarray(stds), np.asarray(times)


def evaluate(fun, keys):
    """Evaluate all metrics of a function."""
    errors, times = zip(*[fun(key) for key in keys])
    return np.asarray(errors), np.asarray(times)


if __name__ == "__main__":
    dim = 10
    num_samples = 2 ** np.arange(4, 12)
    num_restarts = 10

    (matfun, jvp, Av, trace, JJ), (k, sample_fun) = problem(dim)

    x = hutch.trace(Av, key=k, sample_fun=sample_fun)
    y = slq.trace_of_matfun_symmetric(matfun, jvp, 5, key=k, sample_fun=sample_fun)
    assert np.allclose(x, trace, atol=1e-1, rtol=1e-1), (x, trace)
    assert np.allclose(y, trace, atol=1e-1, rtol=1e-1), (y, trace)

    error_fun = func.partial(benchmark_util.rmse_relative, expected=trace)

    @func.partial(benchmark_util.error_and_time, error_fun=error_fun)
    @func.partial(func.jit, static_argnums=0)
    def matvec(num, key):
        """Matrix-vector mult."""
        return hutch.trace(
            Av, key=key, sample_fun=sample_fun, num_samples_per_batch=num
        )

    @func.partial(benchmark_util.error_and_time, error_fun=error_fun)
    @func.partial(func.jit, static_argnums=0)
    def slq_low(num, key):
        """SLQ(1)"""  # noqa: D400,D415
        return slq.trace_of_matfun_symmetric(
            matfun,
            jvp,
            1,
            key=key,
            sample_fun=sample_fun,
            num_samples_per_batch=num,
        )

    @func.partial(benchmark_util.error_and_time, error_fun=error_fun)
    @func.partial(func.jit, static_argnums=0)
    def slq_high(num, key):
        """SLQ(5)"""  # noqa: D400,D415
        return slq.trace_of_matfun_symmetric(
            matfun,
            jvp,
            5,
            key=key,
            sample_fun=sample_fun,
            num_samples_per_batch=num,
        )

    test_keys = prng.split(k, num=num_restarts)
    fig, axes = plt.subplots(ncols=2, tight_layout=True, figsize=(10, 5), dpi=100)
    ax_wp, ax_sparsity = axes

    ax_wp.set_title("Estimating trace(A^2)")
    for fun in [matvec, slq_low, slq_high]:
        errors, stds, times = evaluate_all(fun, progressbar(num_samples), test_keys)
        ax_wp.loglog(
            times,
            errors,
            "o-",
            markeredgecolor="k",
            label=fun.__doc__,
        )
        ax_wp.fill_between(times, errors - stds, errors + stds, alpha=0.25)

    ax_wp.set_ylabel(f"Relative error (mean & std, {num_restarts} restarts)")
    ax_wp.set_xlabel(f"Wall clock time (minimum, {num_restarts} restarts)")
    ax_wp.legend()

    ax_sparsity.set_title("Sparsity pattern of A^2")
    ax_sparsity.spy(JJ)
    ax_sparsity.set_xticks(())
    ax_sparsity.set_yticks(())
    plt.show()



"""Tests for GKL bidiagonalisation."""

from matfree import decomp, test_util
from matfree.backend import linalg, np, prng, testing


@testing.fixture()
def A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 10.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


@testing.parametrize("nrows", [50])
@testing.parametrize("ncols", [49])
@testing.parametrize("num_significant_singular_vals", [4])
@testing.parametrize("order", [6])  # ~1.5 * num_significant_eigvals
def test_golub_kahan_lanczos_bidiagonal(A, order):
    """Test that Lanczos tridiagonalisation yields an orthogonal-tridiagonal decomp."""
    nrows, ncols = np.shape(A)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(ncols,))
    alg = decomp.golub_kahan_lanczos_bidiagonal(order, matrix_shape=np.shape(A))

    def Av(v):
        return A @ v

    def vA(v):
        return v @ A

    Us, Bs, Vs, (b, v) = decomp.decompose_fori_loop(0, order + 1, v0, Av, vA, alg=alg)
    (d_m, e_m) = Bs

    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}

    assert np.shape(Us) == (nrows, order + 1)
    assert np.allclose(Us.T @ Us, np.eye(order + 1), **tols_decomp), Us.T @ Us

    assert np.shape(Vs) == (order + 1, ncols)
    assert np.allclose(Vs @ Vs.T, np.eye(order + 1), **tols_decomp), Vs @ Vs.T

    UAVt = Us.T @ A @ Vs.T
    assert np.allclose(linalg.diagonal(UAVt), d_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(UAVt, 1), e_m, **tols_decomp)

    B = _bidiagonal_dense(d_m, e_m)
    assert np.shape(B) == (order + 1, order + 1)
    assert np.allclose(UAVt, B, **tols_decomp)

    em = np.eye(order + 1)[:, -1]
    AVt = A @ Vs.T
    UtB = Us @ B
    AtUt = A.T @ Us
    VtBtb_plus_bve = Vs.T @ B.T + b * v[:, None] @ em[None, :]
    assert np.allclose(AVt, UtB, **tols_decomp)
    assert np.allclose(AtUt, VtBtb_plus_bve, **tols_decomp)


def _bidiagonal_dense(d, e):
    diag = linalg.diagonal_matrix(d)
    offdiag = linalg.diagonal_matrix(e, 1)
    return diag + offdiag



"""Tests for SVD functionality."""


from matfree import decomp, test_util
from matfree.backend import linalg, np, testing


@testing.fixture()
@testing.parametrize("nrows", [10])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_significant_singular_vals", [3])
def A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 10.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


def test_max_depth(A):
    """The output of full-depth SVD should be equal (*) to linalg.svd().

    (*) Note: The singular values should be identical,
    and the orthogonal matrices should be orthogonal. They are not unique.
    """
    nrows, ncols = np.shape(A)
    depth = min(nrows, ncols) - 1

    def Av(v):
        return A @ v

    def vA(v):
        return v @ A

    v0 = np.ones((ncols,))
    U, S, Vt = decomp.svd(v0, depth, Av, vA, matrix_shape=np.shape(A))
    U_, S_, Vt_ = linalg.svd(A, full_matrices=False)

    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}
    assert np.allclose(S, S_)  # strict "allclose"

    assert np.shape(U) == np.shape(U_)
    assert np.shape(Vt) == np.shape(Vt_)
    assert np.allclose(U @ U.T, U_ @ U_.T, **tols_decomp)
    assert np.allclose(Vt @ Vt.T, Vt_ @ Vt_.T, **tols_decomp)



"""Tests for basic trace estimators."""

from matfree import hutch, montecarlo
from matfree.backend import func, linalg, np, prng, testing


@testing.fixture(name="fun")
def fixture_fun():
    """Create a nonlinear, to-be-differentiated function."""

    def f(x):
        return np.sin(np.flip(np.cos(x)) + 1.0) * np.sin(x) + 1.0

    return f


@testing.fixture(name="key")
def fixture_key():
    """Fix a pseudo-random number generator."""
    return prng.prng_key(seed=1)


@testing.parametrize("num_batches", [1_000])
@testing.parametrize("num_samples_per_batch", [1_000])
@testing.parametrize("dim", [1, 10])
@testing.parametrize("sample_fun", [montecarlo.normal, montecarlo.rademacher])
def test_frobeniusnorm_squared(
    fun, key, num_batches, num_samples_per_batch, dim, sample_fun
):
    """Assert that the Frobenius norm estimate is accurate."""
    # Linearise function
    x0 = prng.uniform(key, shape=(dim,))  # random lin. point
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    # Estimate the trace
    fun = sample_fun(shape=np.shape(x0), dtype=np.dtype(x0))
    estimate = hutch.frobeniusnorm_squared(
        jvp,
        num_batches=num_batches,
        key=key,
        num_samples_per_batch=num_samples_per_batch,
        sample_fun=fun,
    )
    truth = linalg.trace(J.T @ J)
    assert np.allclose(estimate, truth, rtol=1e-2)
